import uuid
import logging
import json
import pathlib

from io import IOBase

import httpx
import pydantic

from .core.exceptions import WappstoError

from .schemas.base import WappstoService
from .schemas import WappstoSchema

from rich.console import Console

console = Console()


class WappstoServiceBuilder:
    """
    Wappsto Service Builder, is a constructor class to build out the request.

    Wappsto Service Builder is a parent class, to be inherited from, for all
    the service, there is, where the only changes are a overload of the
    'Execution' methods, to set the right return types for each one.

    NOTE: https://mail.python.org/pipermail/python-dev/2003-October/038855.html
    """

    def __init__(self, wappsto_service: WappstoService, rest_ref: 'WappstoRest'):
        """
        Initialization of the Wappsto Service constructor.

        Args:
            wappsto_service: The Wappsto Service that should connect to.
            rest_ref: Ref to the Base Rest class.
        """
        self.log = logging.getLogger(__name__)
        self.log.addHandler(logging.NullHandler())

        self._admin: bool = False
        self._file: pathlib.Path | IOBase | None = None
        self._filters: list[str] = []
        self._headers: dict[str, str] = {
            "Accept": "application/json",
            "Content-type": "application/json",
        }
        self._options: dict[str, str | int | float] = {}
        self._parent: dict[uuid.UUID, WappstoService] = {}
        self._service_id: uuid.UUID | None = None
        self.rest_ref: 'WappstoRest' = rest_ref
        self._service: WappstoService = wappsto_service

        self._set_session_to_default()

    def _build_url(
        self,
        service_uuid: uuid.UUID | None = None,
        subservice: WappstoService | None = None,
    ) -> str:
        _vr = self.rest_ref.vr
        url = self.rest_ref.service_url
        service_path: str = "/"
        _parent_id, _parent_service = list(self._parent.items())[0] if self._parent else (None, None)
        if _parent_id:
            service_path += f"{_parent_service.value}/{str(_parent_id)}/"

        service_path += f"{self._service.value}"

        if self._service_id:
            service_path += f"/{str(self._service_id)}"

        if not self._admin:
            return f"https://{url}/{_vr.value}{service_path}"
        return f"https://admin.{url}/{_vr.value}{service_path}"

    def _build_filters(self) -> str:
        the_great_filter: str = "&".join(
            map(lambda x: str(x), self._filters)
        )
        return "" if not the_great_filter else "?" + the_great_filter

    def _set_session_to_default(self):
        if self.rest_ref.default_session is not None:
            self._headers["x-session"] = str(self.rest_ref.default_session)

    def filter(self, filter: str) -> 'WappstoServiceBuilder':
        """
        Applying a filter the the service request.

        Args:
            filter: The filter Criteria to be applied.

        Returns:
            Reference to self.
        """
        self._filters.append(filter)
        return self

    def header(self, **kwargs: str | int | float) -> 'WappstoServiceBuilder':
        """
        Applying a header to the request.

        Args:
            **kwargs: The Header key and value pair.

        Returns:
            Reference to self.
        """
        self._headers.update({
            key: str(value) for key, value in kwargs.items()
        })
        return self

    def option(self, **kwargs: str | int | float) -> 'WappstoServiceBuilder':
        """
        Set a parameter on the request.

        Args:
            **kwargs: The Option key and value pair.

        Returns:
            Reference to self.
        """
        self._options.update(kwargs)
        return self

    def verbose(self) -> 'WappstoServiceBuilder':
        """
        Set the verbose parameter on the request.

        Returns:
            Reference to self.
        """
        self._options.update({'verbose': True})
        return self

    def admin(self) -> 'WappstoServiceBuilder':
        """
        Set the Request to be a admin based request.

        Returns:
            Reference to self.
        """
        self._admin = True
        if self.rest_ref.admin_session is not None:
            self._headers["x-session"] = str(self.rest_ref.admin_session)
        return self

    def parent(
        self,
        parent_id: uuid.UUID,
        parent_service: WappstoService,
    ) -> 'WappstoServiceBuilder':
        """
        Set the parent for the current Wappsto Service.

        Args:
            parent_id: The UUID of the parent.
            parent_service: The Service of the parent.

        Returns:
            Reference to self.
        """
        self._parent[parent_id] = parent_service
        return self

    def service_id(self, service_id: uuid.UUID) -> 'WappstoServiceBuilder':
        """
        Set the UUID of the current service.

        Args:
            service_id: The service UUID for the given service.

        Returns:
            Reference to self.
        """
        self._service_id = service_id
        return self

    def upload_file(self, file_obj: pathlib.Path | IOBase) -> 'WappstoServiceBuilder':
        """
        Set a File to be uploaded.

        Args:
            file_obj: The path to the file, or a file object.

        Returns:
            Reference to self.
        """
        self._file = file_obj
        return self

    """The Execution of the Request."""

    def _ready_data(self) -> str:
        """
        Ready all the data needed for a request.

        This function readies:
         - URL
         - Headers
         - Parameter
         - Options
         - Filters
        """
        the_url: str = self._build_url()
        the_url += self._build_filters()
        return the_url

    def _exception_builder(self, rdata: httpx.Response):
        """
        Check which level the problem came and raise the right Exception.

        Raise:
            WappstoError: Wappsto Problems.
            ConnectionError: Network Problems.
        """
        try:
            the_data = rdata.json()
        except json.JSONDecodeError:
            raise ConnectionError("An error happened doing request.")
        else:
            raise WappstoError(
                code=the_data.get('code'),
                msg=the_data.get('message'),
                url=str(rdata.url),
                http_code=rdata.status_code
            )

    def read(self, expand: int | None = None) -> dict:
        """
        Building and sending the HTTP GET request.

        Args:
            expand: (Optional) If the data returned should be expanded.

        Raise:
            WappstoError: Wappsto Problems.
            ConnectionError: Network Problems.
            httpx.HTTPError  (TODO: FIXME!)
        """
        options = {**self._options}
        if expand is not None:
            options['expand'] = expand
        url: str = self._ready_data()
        rdata = httpx.get(
            url=url,
            headers=self._headers,
            params=options,
        )

        full_url = rdata.url

        if rdata.status_code >= 300:
            console.log(f"service:{self._service.value} -> get", log_locals=True)
            self._exception_builder(rdata)

        r_data = rdata.json()

        if logging.root.level <= logging.INFO:
            console.log(f"service:{self._service.value} -> get", log_locals=True)

        return r_data

    def delete(self) -> dict:
        """
        Building and sending the HTTP DELETE request.

        Raise:
            WappstoError: Wappsto Problems.
            ConnectionError: Network Problems.
            httpx.HTTPError  (TODO: FIXME!)
        """
        url: str = self._ready_data()
        rdata = httpx.delete(
            url=url,
            headers=self._headers,
            params=self._options,
        )

        full_url = rdata.url

        if rdata.status_code >= 300:
            console.log(f"service:{self._service.value} -> delete", log_locals=True)
            self._exception_builder(rdata)

        r_data = rdata.json()

        if logging.root.level <= logging.INFO:
            console.log(f"service:{self._service.value} -> delete", log_locals=True)

        return r_data

    def create(self, data: dict | str) -> dict:
        """
        Building and sending the HTTP POST request.

        Args:
            data: The data object to be created.

        Raise:
            WappstoError: Wappsto Problems.
            ConnectionError: Network Problems.
            httpx.HTTPError  (TODO: FIXME!)
        """
        print(self._headers)
        url: str = self._ready_data()
        rdata = httpx.post(
            url=url,
            headers=self._headers,
            params=self._options,
            json=data if isinstance(data, dict) else None,
            content=data.encode() if isinstance(data, str) else None,
        )

        full_url = rdata.url

        if rdata.status_code >= 300:
            console.log(f"service:{self._service.value} -> create", log_locals=True)
            self._exception_builder(rdata)

        r_data = rdata.json()

        if logging.root.level <= logging.INFO:
            console.log(f"service:{self._service.value} -> create", log_locals=True)

        return r_data

    def update(self, data: dict | str, overwrite: bool = False) -> dict:
        """
        Building and sending the HTTP PUT request.

        Args:
            data: The data object to be updated with.

        Raise:
            WappstoError: Wappsto Problems.
            ConnectionError: Network Problems.
            httpx.HTTPError  (TODO: FIXME!)
        """
        url: str = self._ready_data()

        httpxupdate = httpx.patch if overwrite is True else httpx.put

        rdata = httpxupdate(
            url=url,
            headers=self._headers,
            params=self._options,
            json=data if isinstance(data, dict) else None,
            content=data.encode() if isinstance(data, str) else None,
        )

        full_url = rdata.url

        if rdata.status_code >= 300:
            console.log(f"service:{self._service.value} -> update", log_locals=True)
            self._exception_builder(rdata)

        r_data = rdata.json()

        if logging.root.level <= logging.INFO:
            console.log(f"service:{self._service.value} -> update", log_locals=True)

        return r_data


class WappstoSessionBuilder(WappstoServiceBuilder):
    def read(
        self, expand: int | None = None
    ) -> WappstoSchema.response.Session | WappstoSchema.response.IdList | WappstoSchema.response.AttributeList | list[WappstoSchema.response.Session]:
        r_data = super().read(expand=expand)
        return pydantic.parse_obj_as(
            WappstoSchema.response.Session | WappstoSchema.response.IdList | WappstoSchema.response.AttributeList | list[WappstoSchema.response.Session],
            r_data
        )

    def delete(self) -> WappstoSchema.response.Session | WappstoSchema.response.DeleteList:
        r_data = super().delete()
        return pydantic.parse_obj_as(
            WappstoSchema.response.Session | WappstoSchema.response.DeleteList,
            r_data
        )
    def create(self, data: WappstoSchema.create.Session) -> WappstoSchema.response.Session | list[WappstoSchema.response.Session]:
        r_data = super().create(data=data.json(exclude_none=True))
        return pydantic.parse_obj_as(
            WappstoSchema.response.Session | list[WappstoSchema.response.Session],
            r_data
        )
    def update(self, data: WappstoSchema.update.Session, overwrite: bool = False) -> WappstoSchema.response.Session:
        r_data = super().update(data=data.json(exclude_none=True), overwrite=overwrite)
        return pydantic.parse_obj_as(
            WappstoSchema.response.Session,
            r_data
        )


class WappstoUserBuilder(WappstoServiceBuilder):
    def read(
        self, expand: int | None = None
    ) -> WappstoSchema.response.User | WappstoSchema.response.IdList | WappstoSchema.response.AttributeList | list[WappstoSchema.response.User]:
        r_data = super().read(expand=expand)
        return pydantic.parse_obj_as(
            WappstoSchema.response.User | WappstoSchema.response.IdList | WappstoSchema.response.AttributeList | list[WappstoSchema.response.User],
            r_data
        )

    def delete(self) -> WappstoSchema.response.User | WappstoSchema.response.DeleteList:
        r_data = super().delete()
        return pydantic.parse_obj_as(
            WappstoSchema.response.User | WappstoSchema.response.DeleteList,
            r_data
        )
    def create(self, data: WappstoSchema.create.User) -> WappstoSchema.response.User | list[WappstoSchema.response.User]:
        r_data = super().create(data=data.json(exclude_none=True))
        return pydantic.parse_obj_as(
            WappstoSchema.response.User | list[WappstoSchema.response.User],
            r_data
        )
    def update(self, data: WappstoSchema.update.User, overwrite: bool = False) -> WappstoSchema.response.User:
        r_data = super().update(data=data.json(exclude_none=True), overwrite=overwrite)
        return pydantic.parse_obj_as(
            WappstoSchema.response.User,
            r_data
        )


class WappstoNetworkBuilder(WappstoServiceBuilder):
    def read(
        self, expand: int | None = None
    ) -> WappstoSchema.response.Network | WappstoSchema.response.IdList | WappstoSchema.response.AttributeList | list[WappstoSchema.response.Network]:
        r_data = super().read(expand=expand)
        return pydantic.parse_obj_as(
            WappstoSchema.response.Network | WappstoSchema.response.IdList | WappstoSchema.response.AttributeList | list[WappstoSchema.response.Network],
            r_data
        )

    def delete(self) -> WappstoSchema.response.Network | WappstoSchema.response.DeleteList:
        r_data = super().delete()
        return pydantic.parse_obj_as(
            WappstoSchema.response.Network | WappstoSchema.response.DeleteList,
            r_data
        )
    def create(self, data: WappstoSchema.create.Network) -> WappstoSchema.response.Network | list[WappstoSchema.response.Network]:
        r_data = super().create(data=data.json(exclude_none=True))
        return pydantic.parse_obj_as(
            WappstoSchema.response.Network | list[WappstoSchema.response.Network],
            r_data
        )
    def update(self, data: WappstoSchema.update.Network, overwrite: bool = False) -> WappstoSchema.response.Network:
        r_data = super().update(data=data.json(exclude_none=True), overwrite=overwrite)
        return pydantic.parse_obj_as(
            WappstoSchema.response.Network,
            r_data
        )


class WappstoDeviceBuilder(WappstoServiceBuilder):
    def read(
        self, expand: int | None = None
    ) -> WappstoSchema.response.Device | WappstoSchema.response.IdList | WappstoSchema.response.AttributeList | list[WappstoSchema.response.Device]:
        r_data = super().read(expand=expand)
        return pydantic.parse_obj_as(
            WappstoSchema.response.Device | WappstoSchema.response.IdList | WappstoSchema.response.AttributeList | list[WappstoSchema.response.Device],
            r_data
        )

    def delete(self) -> WappstoSchema.response.Device | WappstoSchema.response.DeleteList:
        r_data = super().delete()
        return pydantic.parse_obj_as(
            WappstoSchema.response.Device | WappstoSchema.response.DeleteList,
            r_data
        )
    def create(self, data: WappstoSchema.create.Device) -> WappstoSchema.response.Device | list[WappstoSchema.response.Device]:
        r_data = super().create(data=data.json(exclude_none=True))
        return pydantic.parse_obj_as(
            WappstoSchema.response.Device | list[WappstoSchema.response.Device],
            r_data
        )
    def update(self, data: WappstoSchema.update.Device, overwrite: bool = False) -> WappstoSchema.response.Device:
        r_data = super().update(data=data.json(exclude_none=True), overwrite=overwrite)
        return pydantic.parse_obj_as(
            WappstoSchema.response.Device,
            r_data
        )


class WappstoValueBuilder(WappstoServiceBuilder):
    def read(
        self, expand: int | None = None
    ) -> WappstoSchema.response.Value | WappstoSchema.response.IdList | WappstoSchema.response.AttributeList | list[WappstoSchema.response.Value]:
        r_data = super().read(expand=expand)
        return pydantic.parse_obj_as(
            WappstoSchema.response.Value | WappstoSchema.response.IdList | WappstoSchema.response.AttributeList | list[WappstoSchema.response.Value],
            r_data
        )

    def delete(self) -> WappstoSchema.response.Value | WappstoSchema.response.DeleteList:
        r_data = super().delete()
        return pydantic.parse_obj_as(
            WappstoSchema.response.Value | WappstoSchema.response.DeleteList,
            r_data
        )
    def create(self, data: WappstoSchema.create.Value) -> WappstoSchema.response.Value | list[WappstoSchema.response.Value]:
        r_data = super().create(data=data.json(exclude_none=True))
        return pydantic.parse_obj_as(
            WappstoSchema.response.Value | list[WappstoSchema.response.Value],
            r_data
        )
    def update(self, data: WappstoSchema.update.Value, overwrite: bool = False) -> WappstoSchema.response.Value:
        r_data = super().update(data=data.json(exclude_none=True), overwrite=overwrite)
        return pydantic.parse_obj_as(
            WappstoSchema.response.Value,
            r_data
        )


class WappstoStateBuilder(WappstoServiceBuilder):
    def read(
        self, expand: int | None = None
    ) -> WappstoSchema.response.State | WappstoSchema.response.IdList | WappstoSchema.response.AttributeList | list[WappstoSchema.response.State]:
        r_data = super().read(expand=expand)
        return pydantic.parse_obj_as(
            WappstoSchema.response.State | WappstoSchema.response.IdList | WappstoSchema.response.AttributeList | list[WappstoSchema.response.State],
            r_data
        )

    def delete(self) -> WappstoSchema.response.State | WappstoSchema.response.DeleteList:
        r_data = super().delete()
        return pydantic.parse_obj_as(
            WappstoSchema.response.State | WappstoSchema.response.DeleteList,
            r_data
        )
    def create(self, data: WappstoSchema.create.State) -> WappstoSchema.response.State | list[WappstoSchema.response.State]:
        r_data = super().create(data=data.json(exclude_none=True))
        return pydantic.parse_obj_as(
            WappstoSchema.response.State | list[WappstoSchema.response.State],
            r_data
        )
    def update(self, data: WappstoSchema.update.State, overwrite: bool = False) -> WappstoSchema.response.State:
        r_data = super().update(data=data.json(exclude_none=True), overwrite=overwrite)
        return pydantic.parse_obj_as(
            WappstoSchema.response.State,
            r_data
        )


# class WappstoUserBuilder(WappstoServiceBuilder):
#     def read(
#         self, expand: int | None = None
#     ) -> WappstoSchema.User: ...
#     def delete(self) -> WappstoSchema.User: ...
#     def create(self, data: WappstoSchema.User) -> WappstoSchema.User: ...
#     def update(self, data: WappstoSchema.User) -> WappstoSchema.User: ...


get_builder = {
    WappstoService.USER: WappstoUserBuilder,
    WappstoService.SESSION: WappstoSessionBuilder,
    WappstoService.NETWORK: WappstoNetworkBuilder,
    WappstoService.DEVICE: WappstoDeviceBuilder,
    WappstoService.VALUE: WappstoValueBuilder,
    WappstoService.STATE: WappstoStateBuilder,
}



# generated by datamodel-codegen:
#   filename:  <stdin>
#   timestamp: 2023-02-17T08:37:02+00:00

from __future__ import annotations

import uuid
from typing import List

from pydantic import BaseModel, Extra, Field, constr


class Meta(BaseModel):
    class Config:
        extra = Extra.allow

    id: uuid.UUID | None = None
    trace: str | None = None
    redirect: constr(regex=r'^[0-9a-zA-Z_-]+$', min_length=1, max_length=200) | None = None
    icon: str | None = None
    tag: list[constr(min_length=2, max_length=20)] | None = None
    tag_by_user: list[constr(min_length=2, max_length=20)] | None = None
    name_by_user: constr(max_length=100) | None = None


class Stream(BaseModel):
    class Config:
        extra = Extra.forbid

    name: str | None = None
    subscription: List[str] = Field(..., min_items=1)
    ignore: list[str] | None = None
    full: bool | None = None
    meta: Meta | None = Field(None, title='meta-2.1')



# generated by datamodel-codegen:
#   filename:  <stdin>
#   timestamp: 2023-02-17T08:37:37+00:00

from __future__ import annotations

from pydantic import BaseModel


class TransferTranslation(BaseModel):
    pass



from typing import Any


class WappstoError(ConnectionError):
    """
    Exception Used if Wappsto reply with an error.
    """
    def __init__(
        self,
        code: int,
        msg: str,
        url: str,
        http_code: int,
        data: Any | None = None
    ) -> None:
        """
        Initialize the WappstoError with the Error Response info.

        Args:
            code: The Wappsto Error code, within the range of -32000 to -32099
            msg: The Wappsto Error message, that shortly describe the error for given code.
            url: The URL that was used to trigger the Exception.
            data: (Optional) The data used when triggered the Exception.
        """
        super().__init__()
        self.code = code
        self.msg = msg
        self.url = url
        self.http_code = http_code
        self.data = data



# generated by datamodel-codegen:
#   filename:  <stdin>
#   timestamp: 2023-02-17T08:37:59+00:00

from __future__ import annotations

import uuid
import datetime
from enum import Enum

from pydantic import BaseModel, Extra, constr


class ProviderItem(BaseModel):
    class Config:
        extra = Extra.forbid

    name: str | None = None
    picture: str | None = None
    type: str | None = None


class PublicEnum(Enum):
    nickname = 'nickname'
    first_name = 'first_name'
    last_name = 'last_name'
    email = 'email'
    phone = 'phone'
    role = 'role'


class StatusEnum(Enum):
    pending = 'pending'
    refused = 'refused'
    accepted = 'accepted'
    send = 'send'
    not_sent = 'not_sent'
    archive = 'archive'


class OtherEmailItem(BaseModel):
    contact: constr(regex=r'^\S+@\S+\.\S+$') | None = None
    status:  StatusEnum | None = None
    last_update: datetime.datetime | None = None


class OtherSm(BaseModel):
    contact: str | None = None
    status:  StatusEnum | None = None
    last_update: datetime.datetime | None = None


class Type(Enum):
    soft = 'soft'
    strong = 'strong'


class BanItem(BaseModel):
    type:  Type | None = None
    begin_ban: datetime.datetime | None = None
    end_ban: datetime.datetime | None = None
    motivation: str | None = None


class MetaItem(BaseModel):
    class Config:
        extra = Extra.allow

    id: uuid.UUID | None = None
    trace: str | None = None
    redirect: constr(regex=r'^[0-9a-zA-Z_-]+$', min_length=1, max_length=200) | None = None
    icon: str | None = None
    tag: list[constr(min_length=2, max_length=20)] | None = None
    tag_by_user: list[constr(min_length=2, max_length=20)] | None = None
    name_by_user: constr(max_length=100) | None = None


class Virtualuser(BaseModel):
    class Config:
        extra = Extra.forbid

    first_name: str | None = None
    last_name: str | None = None
    email: constr(regex=r'^\S+@\S+\.\S+$') | None = None
    phone: str | None = None
    name: str | None = None
    role: str | None = None
    nickname: str | None = None
    language: str | None = None
    friend: bool | None = None
    blocked: bool | None = None
    provider: list[ProviderItem] | None = None
    public: list[PublicEnum] | None = None
    verified_email: bool | None = None
    verified_sms: bool | None = None
    other_email: list[OtherEmailItem] | None = None
    other_sms: list[OtherSm] | None = None
    admin: bool | None = None
    founder: bool | None = None
    ban:  BanItem | None = None
    meta:  MetaItem | None = None



# generated by datamodel-codegen:
#   filename:  <stdin>
#   timestamp: 2023-02-17T08:27:23+00:00

from __future__ import annotations

import uuid
import datetime
from enum import Enum
from typing import Any, Literal

from pydantic import BaseModel, Extra, Field, conint, constr


class Assigned(Enum):
    unassigned = 'unassigned'
    admin = 'admin'


class WarningItem(BaseModel):
    message: str | None = None
    data: dict[str, Any] | None = None
    code: int | None = None


class Meta(BaseModel):
    class Config:
        extra = Extra.forbid

    id: uuid.UUID | None = None
    type: str | None = None
    version: constr(regex=r'(\d+)\.(\d+)') | None = None
    manufacturer: uuid.UUID | Assigned | None = None
    owner: uuid.UUID | Assigned | None = None
    created: datetime.datetime | Literal['not_available'] | None = None
    updated: datetime.datetime | Literal['not_available'] | None = None
    changed: datetime.datetime | Literal['not_available'] | None = None
    application: uuid.UUID | None = None
    revision: int | None = None
    trace: str | None = None
    oem: str | None = None
    deprecated: bool | None = None
    redirect: constr(regex=r'^[0-9a-zA-Z_-]+$', min_length=1, max_length=200) | None = None
    size: int | None = None
    path: str | None = None
    parent: uuid.UUID | None = None
    error: dict[str, Any] | None = None
    icon: str | None = None
    tag: list[constr(min_length=2, max_length=20)] | None = None
    tag_by_user: list[constr(min_length=2, max_length=20)] | None = None
    name_by_user: constr(max_length=100) | None = None
    warning: list[WarningItem] | None = None
    parent_name: dict[str, Any] | None = None
    read_only: bool | None = None
    original_id: uuid.UUID | None = None
    alert: list[uuid.UUID] | None = None
    contract: list[uuid.UUID] | None = None


class Type(Enum):
    one_time = 'one_time'
    subscription = 'subscription'


class Interval(Enum):
    day = 'day'
    week = 'week'
    month = 'month'
    year = 'year'


class Recurring(BaseModel):
    class Config:
        extra = Extra.forbid

    interval: Interval
    interval_count: conint(ge=1)


class PricesWithTax(BaseModel):
    amount: int | None = None
    prepaid_extra_cost: int | None = None
    topup_extra_cost: int | None = None


class PricesWithoutTax(BaseModel):
    amount: int | None = None
    prepaid_extra_cost: int | None = None
    topup_extra_cost: int | None = None


class Tax(BaseModel):
    id: uuid.UUID | None = None
    prices_with_tax: PricesWithTax | None = None
    prices_without_tax: PricesWithoutTax | None = None
    name: str | None = None
    description: str | None = None
    jurisdiction: str | None = None
    inclusive: bool | None = None
    percentage: int | None = None
    state: str | None = None


class ApplicationProduct(BaseModel):
    class Config:
        extra = Extra.forbid

    meta: Meta | None = Field(None, title='meta-2.1')
    name: str | None = None
    description: str | None = None
    active: bool | None = None
    deleted: bool | None = None
    stripe: dict[str, Any] | None = None
    type:  Type | None = None
    currency: str | None = None
    amount: conint(ge=1) | None = None
    recurring: Recurring | None = None
    tax: Tax | None = None
    trial_period_days: conint(ge=1) | None = None



# generated by datamodel-codegen:
#   filename:  <stdin>
#   timestamp: 2023-02-17T08:36:32+00:00

from __future__ import annotations

import uuid
import datetime
from enum import Enum

from pydantic import BaseModel, Extra, Field, constr


class Type(Enum):
    Report = 'Report'
    Control = 'Control'


class Status(Enum):
    Send = 'Send'
    Pending = 'Pending'
    Failed = 'Failed'
    Queued = 'Queued'


class Meta(BaseModel):
    class Config:
        extra = Extra.allow

    id: uuid.UUID | None = None
    trace: str | None = None
    redirect: constr(regex=r'^[0-9a-zA-Z_-]+$', min_length=1, max_length=200) | None = None
    icon: str | None = None
    tag: list[constr(min_length=2, max_length=20)] | None = None
    tag_by_user: list[constr(min_length=2, max_length=20)] | None = None
    name_by_user: constr(max_length=100) | None = None
    historical: bool | None = None


class State(BaseModel):
    class Config:
        extra = Extra.allow

    timestamp: datetime.datetime | None = None
    data: str
    type: Type
    status: Status | None = None
    meta: Meta | None = Field(None, title='meta-2.1')



# generated by datamodel-codegen:
#   filename:  <stdin>
#   timestamp: 2023-02-17T08:37:16+00:00

from __future__ import annotations

import uuid
from enum import Enum

from pydantic import BaseModel, Extra, conint, constr


class MetaItem(BaseModel):
    class Config:
        extra = Extra.allow

    id: uuid.UUID | None = None
    trace: str | None = None
    redirect: constr(regex=r'^[0-9a-zA-Z_-]+$', min_length=1, max_length=200) | None = None
    icon: str | None = None
    tag: list[constr(min_length=2, max_length=20)] | None = None
    tag_by_user: list[constr(min_length=2, max_length=20)] | None = None
    name_by_user: constr(max_length=100) | None = None


class UserDailyLimitItem(BaseModel):
    class Config:
        extra = Extra.forbid

    point: conint(ge=0) | None = None
    document: conint(ge=0) | None = None
    log_row: conint(ge=0) | None = None
    traffic: conint(ge=0) | None = None
    iot_traffic: conint(ge=0) | None = None
    stream_traffic: conint(ge=0) | None = None
    file: conint(ge=0) | None = None
    request: conint(ge=0) | None = None
    request_time: conint(ge=0) | None = None


class NetworkDailyLimitItem(BaseModel):
    class Config:
        extra = Extra.forbid

    point: conint(ge=0) | None = None
    iot_traffic: conint(ge=0) | None = None
    request: conint(ge=0) | None = None
    request_time: conint(ge=0) | None = None


class DynamicLimitEnum(Enum):
    dynamic = 'dynamic'


class DynamicLimit(BaseModel):
    __root__: conint(ge=0) | DynamicLimitEnum


class PointManagementItem(BaseModel):
    base_point: DynamicLimit | None = None
    base_network: DynamicLimit | None = None
    user_daily_limit:  UserDailyLimitItem | None = None
    network_daily_limit:  NetworkDailyLimitItem | None = None


class Subuser(BaseModel):
    class Config:
        extra = Extra.allow

    first_name: str | None = None
    last_name: str | None = None
    email: constr(regex=r'^\S+@\S+\.\S+$') | None = None
    phone: str | None = None
    name: str | None = None
    role: str | None = None
    nickname: str | None = None
    language: str | None = None
    login_username: constr(regex=r'^[0-9a-zA-Z_.]{1,40}$') | None = None
    login_password: str | None = None
    meta:  MetaItem | None = None
    point_management:  PointManagementItem | None = None



# generated by datamodel-codegen:
#   filename:  <stdin>
#   timestamp: 2023-02-17T08:33:56+00:00

from __future__ import annotations

import uuid
from enum import Enum
from typing import Any

from pydantic import BaseModel, Extra, constr


class ApplicationItem(BaseModel):
    class Config:
        extra = Extra.forbid

    sharing: bool | None = None
    permitted: list[uuid.UUID] | None = None


class MethodItem(BaseModel):
    create: bool | None = None
    update: bool | None = None
    retrieve: bool | None = None
    delete: bool | None = None


class StateItem(BaseModel):
    class Config:
        extra = Extra.forbid

    min_state: float | None = None
    max_state: float | None = None


class InstallationItem(BaseModel):
    class Config:
        extra = Extra.forbid

    shareable: bool | None = None


class FileItem(BaseModel):
    background: bool | None = None


class AclAttribute(Enum):
    parent_name = 'parent_name'


class RestrictionItem(BaseModel):
    class Config:
        extra = Extra.forbid

    child: str | None = None
    create: list[str] | None = None
    application:  ApplicationItem | None = None
    method:  MethodItem | None = None
    time: dict[str, Any] | None = None
    state:  StateItem | None = None
    installation:  InstallationItem | None = None
    file:  FileItem | None = None
    acl_attributes: list[AclAttribute] | None = None


class IdEnum(Enum):
    global_ = 'global'
    parent_owner = 'parent_owner'
    parent_manufacturer = 'parent_manufacturer'


class MetaItem(BaseModel):
    class Config:
        extra = Extra.allow

    id: uuid.UUID | constr(
        regex=r'(?:[a-zA-Z0-9!#$%&\'*+/=?^_`{|}~-]+(?:\.[a-zA-Z0-9!#$%&\'*+/=?^_`{|}~-]+)*|"(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21\x23-\x5b\x5d-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])*")@(?:(?:[a-zA-Z0-9](?:[a-zA-Z0-9-]*[a-zA-Z0-9])?\.)+[a-zA-Z0-9](?:[a-zA-Z0-9-]*[a-zA-Z0-9])?|\[(?:(?:(2(5[0-5]|[0-4][0-9])|1[0-9][0-9]|[1-9]?[0-9]))\.){3}(?:(2(5[0-5]|[0-4][0-9])|1[0-9][0-9]|[1-9]?[0-9])|[a-zA-Z0-9-]*[a-zA-Z0-9]:(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21-\x5a\x53-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])+)\])'
    ) | IdEnum | None = None
    trace: str | None = None
    redirect: constr(regex=r'^[0-9a-zA-Z_-]+$', min_length=1, max_length=200) | None = None
    icon: str | None = None
    tag: list[constr(min_length=2, max_length=20)] | None = None
    tag_by_user: list[constr(min_length=2, max_length=20)] | None = None
    name_by_user: constr(max_length=100) | None = None


class Permission(BaseModel):
    class Config:
        extra = Extra.forbid

    name: str | None = None
    message: str | None = None
    propagate: bool | None = None
    restriction: list[RestrictionItem] | None = None
    meta:  MetaItem | None = None



# generated by datamodel-codegen:
#   filename:  <stdin>
#   timestamp: 2023-02-17T08:28:49+00:00

from __future__ import annotations

import uuid
import datetime
from enum import Enum
from typing import Any, Literal

from pydantic import BaseModel, Extra, Field, confloat, constr


class Assigned(Enum):
    unassigned = 'unassigned'
    admin = 'admin'


class WarningItem(BaseModel):
    message: str | None = None
    data: dict[str, Any] | None = None
    code: int | None = None


class Meta(BaseModel):
    class Config:
        extra = Extra.forbid

    id: uuid.UUID | None = None
    type: str | None = None
    version: constr(regex=r'(\d+)\.(\d+)') | None = None
    manufacturer: uuid.UUID | Assigned | None = None
    owner: uuid.UUID | Assigned | None = None
    created: datetime.datetime | Literal['not_available'] | None = None
    updated: datetime.datetime | Literal['not_available'] | None = None
    changed: datetime.datetime | Literal['not_available'] | None = None
    application: uuid.UUID | None = None
    revision: int | None = None
    trace: str | None = None
    oem: str | None = None
    deprecated: bool | None = None
    redirect: constr(regex=r'^[0-9a-zA-Z_-]+$', min_length=1, max_length=200) | None = None
    size: int | None = None
    path: str | None = None
    parent: uuid.UUID | None = None
    error: dict[str, Any] | None = None
    icon: str | None = None
    tag: list[constr(min_length=2, max_length=20)] | None = None
    tag_by_user: list[constr(min_length=2, max_length=20)] | None = None
    name_by_user: constr(max_length=100) | None = None
    warning: list[WarningItem] | None = None
    parent_name: dict[str, Any] | None = None
    read_only: bool | None = None
    original_id: uuid.UUID | None = None
    alert: list[uuid.UUID] | None = None
    contract: list[uuid.UUID] | None = None


class Recurrency(Enum):
    month = 'month'
    quarter = 'quarter'
    year = 'year'


class Language(Enum):
    en = 'en'
    dk = 'dk'


class Information(BaseModel):
    company: str | None = None
    street: str | None = None
    post_number: str | None = None
    city: str | None = None
    country: str | None = None
    tax_code: str | None = None
    tax_code_type: str | None = None
    email: str | None = None
    phone: str | None = None


class TheirReference(BaseModel):
    name: str | None = None
    email: str | None = None
    phone: str | None = None


class OurReference(BaseModel):
    name: str | None = None
    email: str | None = None
    phone: str | None = None


class Meta1(BaseModel):
    class Config:
        extra = Extra.forbid

    id: uuid.UUID | None = None
    type: str | None = None
    version: constr(regex=r'(\d+)\.(\d+)') | None = None
    manufacturer: uuid.UUID | Assigned | None = None
    owner: uuid.UUID | Assigned | None = None
    created: datetime.datetime | Literal['not_available'] | None = None
    updated: datetime.datetime | Literal['not_available'] | None = None
    changed: datetime.datetime | Literal['not_available'] | None = None
    application: uuid.UUID | None = None
    revision: int | None = None
    trace: str | None = None
    oem: str | None = None
    deprecated: bool | None = None
    redirect: constr(regex=r'^[0-9a-zA-Z_-]+$', min_length=1, max_length=200) | None = None
    size: int | None = None
    path: str | None = None
    parent: uuid.UUID | None = None
    error: dict[str, Any] | None = None
    icon: str | None = None
    tag: list[constr(min_length=2, max_length=20)] | None = None
    tag_by_user: list[constr(min_length=2, max_length=20)] | None = None
    name_by_user: constr(max_length=100) | None = None
    warning: list[WarningItem] | None = None
    parent_name: dict[str, Any] | None = None
    read_only: bool | None = None
    original_id: uuid.UUID | None = None
    alert: list[uuid.UUID] | None = None
    contract: list[uuid.UUID] | None = None


class Recurrency1(Enum):
    one_time = 'one_time'
    subscription = 'subscription'


class Type(Enum):
    network = 'network'
    installation = 'installation'
    application = 'application'
    creator = 'creator'
    generic = 'generic'


class InvoiceItem(BaseModel):
    class Config:
        extra = Extra.forbid

    recurrency:  Recurrency1 | None = None
    type: Type
    filter: str | None = None
    filter_by_name: str | None = None
    filter_by_application: uuid.UUID | None = None
    quantity: int | None = None
    cost: int
    description: str
    dynamic: bool | None = None
    unit: str | None = None
    total: int | None = None


class CustomerInvoiceItem(BaseModel):
    class Config:
        extra = Extra.forbid

    meta: Meta1 | None = Field(None, title='meta-2.1')
    invoice_number: int | None = None
    name: str | None = None
    description: str | None = None
    paid: bool | None = None
    sent: bool | None = None
    overdue: bool | None = None
    total: int | None = None
    subtotal: int | None = None
    tax_total: int | None = None
    discount_total: int | None = None
    currency: str | None = None
    archived: bool | None = None
    due_days_after_sending_invoice: int | None = None
    create_automatic_next_invoice: bool | None = None
    due_timestamp: datetime.datetime | None = None
    paid_timestamp: datetime.datetime | None = None
    sent_timestamp: datetime.datetime | None = None
    due_payment_timestamp: datetime.datetime | None = None
    discount: confloat(ge=0.0, le=100.0) | None = None
    discount_before_tax: bool | None = None
    tax: confloat(ge=0.0, le=100.0) | None = None
    tax_included: bool | None = None
    note: list[str] | None = None
    invoice: list[InvoiceItem] | None = None


class Customer(BaseModel):
    class Config:
        extra = Extra.forbid

    meta: Meta | None = Field(None, title='meta-2.1')
    user_id: str
    seller: uuid.UUID | None = None
    currency: str | None = None
    due_days_after_sending_invoice: int | None = None
    discount_before_tax: bool | None = None
    tax_included: bool | None = None
    note: list[str] | None = None
    next_invoice: dict[str, Any] | None = None
    last_invoice: dict[str, Any] | None = None
    recurrency:  Recurrency | None = None
    create_automatic_next_invoice: bool | None = None
    archived: bool | None = None
    user_exists: bool | None = None
    language:  Language | None = None
    information:  Information | None = None
    their_reference:  TheirReference | None = None
    our_reference:  OurReference | None = None
    customer_number: int | None = None
    customer_invoice: list[CustomerInvoiceItem | uuid.UUID] | None = None



# generated by datamodel-codegen:
#   filename:  <stdin>
#   timestamp: 2023-02-17T08:28:56+00:00

from __future__ import annotations

import uuid
import datetime
from enum import Enum

from pydantic import BaseModel, Extra, Field, confloat, constr


class Meta(BaseModel):
    class Config:
        extra = Extra.allow

    id: uuid.UUID | None = None
    trace: str | None = None
    redirect: constr(regex=r'^[0-9a-zA-Z_-]+$', min_length=1, max_length=200) | None = None
    icon: str | None = None
    tag: list[constr(min_length=2, max_length=20)] | None = None
    tag_by_user: list[constr(min_length=2, max_length=20)] | None = None
    name_by_user: constr(max_length=100) | None = None


class Recurrency(Enum):
    one_time = 'one_time'
    subscription = 'subscription'


class Type(Enum):
    network = 'network'
    installation = 'installation'
    application = 'application'
    creator = 'creator'
    generic = 'generic'


class InvoiceItem(BaseModel):
    class Config:
        extra = Extra.allow

    recurrency:  Recurrency | None = None
    type: Type
    filter: str | None = None
    filter_by_name: str | None = None
    filter_by_application: uuid.UUID | None = None
    quantity: int | None = None
    cost: int
    description: str
    dynamic: bool | None = None
    unit: str | None = None


class CustomerInvoice(BaseModel):
    class Config:
        extra = Extra.allow

    meta: Meta | None = Field(None, title='meta-2.1')
    name: str | None = None
    description: str | None = None
    paid: bool | None = None
    sent: bool | None = None
    currency: str | None = None
    archived: bool | None = None
    due_days_after_sending_invoice: int | None = None
    due_timestamp: datetime.datetime | None = None
    paid_timestamp: datetime.datetime | None = None
    sent_timestamp: datetime.datetime | None = None
    due_payment_timestamp: datetime.datetime | None = None
    discount: confloat(ge=0.0, le=100.0) | None = None
    discount_before_tax: bool | None = None
    create_automatic_next_invoice: bool | None = None
    tax: confloat(ge=0.0, le=100.0) | None = None
    tax_included: bool | None = None
    note: list[str] | None = None
    invoice: list[InvoiceItem] | None = None



# generated by datamodel-codegen:
#   filename:  <stdin>
#   timestamp: 2023-02-17T08:30:31+00:00

from __future__ import annotations

import uuid
import datetime
from enum import Enum
from typing import Any, Literal

from pydantic import BaseModel, Extra, Field, constr


class Assigned(Enum):
    unassigned = 'unassigned'
    admin = 'admin'


class WarningItem(BaseModel):
    message: str | None = None
    data: dict[str, Any] | None = None
    code: int | None = None


class Meta(BaseModel):
    class Config:
        extra = Extra.forbid

    id: uuid.UUID | None = None
    type: str | None = None
    version: constr(regex=r'(\d+)\.(\d+)') | None = None
    manufacturer: uuid.UUID | Assigned | None = None
    owner: uuid.UUID | Assigned | None = None
    created: datetime.datetime | Literal['not_available'] | None = None
    updated: datetime.datetime | Literal['not_available'] | None = None
    changed: datetime.datetime | Literal['not_available'] | None = None
    application: uuid.UUID | None = None
    revision: int | None = None
    trace: str | None = None
    oem: str | None = None
    deprecated: bool | None = None
    redirect: constr(regex=r'^[0-9a-zA-Z_-]+$', min_length=1, max_length=200) | None = None
    size: int | None = None
    path: str | None = None
    parent: uuid.UUID | None = None
    error: dict[str, Any] | None = None
    icon: str | None = None
    tag: list[constr(min_length=2, max_length=20)] | None = None
    tag_by_user: list[constr(min_length=2, max_length=20)] | None = None
    name_by_user: constr(max_length=100) | None = None
    warning: list[WarningItem] | None = None
    parent_name: dict[str, Any] | None = None
    read_only: bool | None = None
    original_id: uuid.UUID | None = None
    alert: list[uuid.UUID] | None = None
    contract: list[uuid.UUID] | None = None


class FailedItem(BaseModel):
    id: uuid.UUID | None = None
    last_update: datetime.datetime | None = None


class ExpectedState(BaseModel):
    class Config:
        extra = Extra.forbid

    meta: Meta | None = Field(None, title='meta-2.1')
    name: str | None = None
    application: uuid.UUID | None = None
    product: str | None = None
    network_name: str | None = None
    device_name: str | None = None
    value_type: str | None = None
    manufacturer: uuid.UUID | None = None
    expected_update: int | None = None
    check_if_network_online: bool | None = None
    check_if_points_are_finished: bool | None = None
    number_failed: int | None = None
    number_states: int | None = None
    failed: list[FailedItem] | None = None



# generated by datamodel-codegen:
#   filename:  <stdin>
#   timestamp: 2023-02-17T08:29:50+00:00

from __future__ import annotations

import uuid
import datetime
from enum import Enum
from typing import Any

from pydantic import BaseModel, Extra, conint, constr


class InclusionStatu(Enum):
    STATUS_DEVICE_INCLUDING = 'STATUS_DEVICE_INCLUDING'
    STATUS_DEVICE_INCLUSION_SUCCESS = 'STATUS_DEVICE_INCLUSION_SUCCESS'
    STATUS_DEVICE_INCLUSION_FAILURE = 'STATUS_DEVICE_INCLUSION_FAILURE'
    STATUS_DEVICE_REPORTING = 'STATUS_DEVICE_REPORTING'
    STATUS_DEVICE_REPORT_SUCCESS = 'STATUS_DEVICE_REPORT_SUCCESS'
    STATUS_DEVICE_REPORT_FAILURE = 'STATUS_DEVICE_REPORT_FAILURE'
    EXCLUDED = 'EXCLUDED'
    INCLUDED = 'INCLUDED'


class FirmwareStatu(Enum):
    UP_TO_DATE = 'UP_TO_DATE'
    UPDATE_AVAILABLE = 'UPDATE_AVAILABLE'
    UPLOADING = 'UPLOADING'
    UPLOAD_COMPLETE = 'UPLOAD_COMPLETE'
    UPLOADING_FAILURE = 'UPLOADING_FAILURE'
    FLASHING = 'FLASHING'
    FLASHING_COMPLETE = 'FLASHING_COMPLETE'
    FLASHING_FAILURE = 'FLASHING_FAILURE'


class CommandEnum(Enum):
    idle = 'idle'
    firmware_upload = 'firmware_upload'
    firmware_flash = 'firmware_flash'
    firmware_cancel = 'firmware_cancel'
    include = 'include'
    exclude = 'exclude'
    connection_check = 'connection_check'


class MetaItem(BaseModel):
    class Config:
        extra = Extra.allow

    id: uuid.UUID | None = None
    trace: str | None = None
    redirect: constr(regex=r'^[0-9a-zA-Z_-]+$', min_length=1, max_length=200) | None = None
    icon: str | None = None
    tag: list[constr(min_length=2, max_length=20)] | None = None
    tag_by_user: list[constr(min_length=2, max_length=20)] | None = None
    name_by_user: constr(max_length=100) | None = None
    historical: bool | None = None


class Level(Enum):
    important = 'important'
    error = 'error'
    warning = 'warning'
    info = 'info'
    debug = 'debug'


class MetaItem1(BaseModel):
    class Config:
        extra = Extra.allow

    id: uuid.UUID | None = None
    trace: str | None = None
    redirect: constr(regex=r'^[0-9a-zA-Z_-]+$', min_length=1, max_length=200) | None = None
    icon: str | None = None
    tag: list[constr(min_length=2, max_length=20)] | None = None
    tag_by_user: list[constr(min_length=2, max_length=20)] | None = None
    name_by_user: constr(max_length=100) | None = None


class StatusItem(BaseModel):
    class Config:
        extra = Extra.allow

    message: str | None = None
    timestamp: datetime.datetime | None = None
    data: str | None = None
    info: dict[str, Any] | None = None
    level:  Level | None = None
    type: str | None = None
    meta:  MetaItem1 | None = None


class PermissionEnum(Enum):
    r = 'r'
    w = 'w'
    rw = 'rw'
    wr = 'wr'
    none = 'none'


class StatusEnum(Enum):
    ok = 'ok'
    update = 'update'
    pending = 'pending'


class NumberItem(BaseModel):
    class Config:
        extra = Extra.forbid

    min: float | constr(regex=r'^[-+]?[0-9]+[.]?[0-9]*([eE][-+]?[0-9]+)?$') | None = None
    max: float | constr(regex=r'^[-+]?[0-9]+[.]?[0-9]*([eE][-+]?[0-9]+)?$') | None = None
    step: float | constr(regex=r'^[-+]?[0-9]+[.]?[0-9]*([eE][-+]?[0-9]+)?$') | None = None
    unit: str | None = None
    si_conversion: str | None = None
    mapping: dict[str, Any] | None = None
    ordered_mapping: bool | None = None
    meaningful_zero: bool | None = None


class StringItem(BaseModel):
    class Config:
        extra = Extra.forbid

    max: conint(ge=1, multiple_of=1) | None = None
    encoding: str | None = None


class BlobItem(BaseModel):
    class Config:
        extra = Extra.forbid

    max: conint(ge=1, multiple_of=1) | None = None
    encoding: str | None = None


class XmlItem(BaseModel):
    class Config:
        extra = Extra.forbid

    xsd: str | None = None
    namespace: str | None = None


class MetaItem2(BaseModel):
    class Config:
        extra = Extra.allow

    id: uuid.UUID | None = None
    trace: str | None = None
    redirect: constr(regex=r'^[0-9a-zA-Z_-]+$', min_length=1, max_length=200) | None = None
    icon: str | None = None
    tag: list[constr(min_length=2, max_length=20)] | None = None
    tag_by_user: list[constr(min_length=2, max_length=20)] | None = None
    name_by_user: constr(max_length=100) | None = None
    historical: bool | None = None


class Type(Enum):
    Report = 'Report'
    Control = 'Control'


class StatusEnum1(Enum):
    Send = 'Send'
    Pending = 'Pending'
    Failed = 'Failed'
    Queued = 'Queued'


class StateItem(BaseModel):
    class Config:
        extra = Extra.allow

    timestamp: datetime.datetime | None = None
    data: str | None = None
    type:  Type | None = None
    status:  StatusEnum1 | None = None
    meta:  MetaItem2 | None = None


class Level1(Enum):
    important = 'important'
    error = 'error'
    success = 'success'
    warning = 'warning'
    info = 'info'
    debug = 'debug'


class MetaItem4(BaseModel):
    class Config:
        extra = Extra.allow

    id: uuid.UUID | None = None
    trace: str | None = None
    redirect: constr(regex=r'^[0-9a-zA-Z_-]+$', min_length=1, max_length=200) | None = None
    icon: str | None = None
    tag: list[constr(min_length=2, max_length=20)] | None = None
    tag_by_user: list[constr(min_length=2, max_length=20)] | None = None
    name_by_user: constr(max_length=100) | None = None


class EventlogItem(BaseModel):
    class Config:
        extra = Extra.allow

    message: str | None = None
    timestamp: datetime.datetime | None = None
    info: dict[str, Any] | None = None
    level:  Level1 | None = None
    type: str | None = None
    meta:  MetaItem4 | None = None


class InfoItem(BaseModel):
    class Config:
        extra = Extra.forbid

    enabled: bool | None = None


class ValueItem(BaseModel):
    class Config:
        extra = Extra.allow

    name: str | None = None
    description: str | None = None
    type: str | None = None
    period: str | None = None
    delta: str | None = None
    permission:  PermissionEnum | None = None
    status:  StatusEnum | None = None
    number:  NumberItem | None = None
    string:  StringItem | None = None
    blob:  BlobItem | None = None
    xml:  XmlItem | None = None
    meta:  MetaItem2 | None = None
    state: list[StateItem | uuid.UUID] | None = None
    eventlog: list[EventlogItem | uuid.UUID] | None = None
    info:  InfoItem | None = None


class Device(BaseModel):
    class Config:
        extra = Extra.allow

    name: str | None = None
    control_timeout: int | None = None
    control_when_offline: bool | None = None
    manufacturer: str | None = None
    product: str | None = None
    version: str | None = None
    serial: str | None = None
    description: str | None = None
    protocol: str | None = None
    communication: str | None = None
    included: str | None = None
    inclusion_status:  InclusionStatu | None = None
    firmware_status:  FirmwareStatu | None = None
    firmware_upload_progress: str | None = None
    firmware_available_version: str | None = None
    command:  CommandEnum | None = None
    meta:  MetaItem | None = None
    status: list[StatusItem | uuid.UUID] | None = None
    value: list[ValueItem | uuid.UUID] | None = None
    info:  InfoItem | None = None



from setuptools import find_packages
from setuptools import setup

import pathlib
import codecs
import os
import re


readme_file = pathlib.Path('README.md')
changelog_file = pathlib.Path('CHANGELOG.md')

here = os.path.abspath(os.path.dirname(__file__))


def get_long_description():
    long_description = ""

    if not readme_file.exists():
        return ""

    with readme_file.open("r") as file:
        long_description += file.read()

    long_description += "\n\n"

    if not changelog_file.exists():
        return long_description

    with changelog_file.open("r") as file:
        long_description += file.read()

    return long_description


def find_version(*file_paths):
    path = os.path.join(here, *file_paths)
    with codecs.open(path, 'r') as fp:
        version_file = fp.read()
        version_match = re.search(r"__version__ = ['\"]([^'\"]*)['\"]",
                                  version_file, re.M)
        if version_match:
            return version_match.group(1)
        raise RuntimeError("Unable to find version string.")


def find_author(*file_paths):
    path = os.path.join(here, *file_paths)
    with codecs.open(path, 'r') as fp:
        auther_file = fp.read()
        auther_match = re.search(r"__auther__ = ['\"]([^'\"]*)['\"]",
                                 auther_file, re.M)
        if auther_match:
            return auther_match.group(1)
        raise RuntimeError("Unable to find auther string.")


setup(
    name="wappstorest",
    python_requires='>=3.9.0',
    version=find_version("src", "wappstorest", "__init__.py"),
    author=find_author("src", "wappstorest", "__init__.py"),
    author_email="support@seluxit.com",
    license="Apache-2.0",
    description="Simple Wappsto Python user-interface to Wappsto Rest",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/Wappsto/python-wappsto-rest",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ],
    # package_data={
    #     'wappstorest': ['WappstoRest.pyi'],
    #     'wappstorest/schemas': ['schemas/jsonrpc.pyi'],
    # },
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    # tests_require=[
    #     'pytest',
    #     'tox'
    # ],
    package_data={'wappstorest': ['../README.md', '../CHANGELOG.md']},
    install_requires=[
        'pydantic==1.9.2',
        'httpx>=0.23.0',
        'websocket-client>=0.59.0',
        'rich>=12.0.0'
    ],
    # entry_points={  # TODO: fix __main__.py to be optional.
    #     "console_scripts": "wappstoiot=wappstoiot:__main__"
    # },
    # extras_require={  # TODO: fix __main__.py to be optional.
    #     "cli": [
    #         'requests',
    #     ]
    # },
)



# generated by datamodel-codegen:
#   filename:  <stdin>
#   timestamp: 2023-02-17T08:37:14+00:00

from __future__ import annotations

import uuid
import datetime
from enum import Enum
from typing import Any, Literal

from pydantic import BaseModel, Extra, Field, conint, constr


class Assigned(Enum):
    unassigned = 'unassigned'
    admin = 'admin'




class WarningItem(BaseModel):
    message: str | None = None
    data: dict[str, Any] | None = None
    code: int | None = None


class Meta(BaseModel):
    class Config:
        extra = Extra.forbid

    id: uuid.UUID | None = None
    type: str | None = None
    version: constr(regex=r'(\d+)\.(\d+)') | None = None
    manufacturer: uuid.UUID | Assigned | None = None
    owner: uuid.UUID | Assigned | None = None
    created: datetime.datetime | Literal['not_available'] | None = None
    updated: datetime.datetime | Literal['not_available'] | None = None
    changed: datetime.datetime | Literal['not_available'] | None = None
    application: uuid.UUID | None = None
    revision: int | None = None
    trace: str | None = None
    oem: str | None = None
    deprecated: bool | None = None
    redirect: constr(regex=r'^[0-9a-zA-Z_-]+$', min_length=1, max_length=200) | None = None
    size: int | None = None
    path: str | None = None
    parent: uuid.UUID | None = None
    error: dict[str, Any] | None = None
    icon: str | None = None
    tag: list[constr(min_length=2, max_length=20)] | None = None
    tag_by_user: list[constr(min_length=2, max_length=20)] | None = None
    name_by_user: constr(max_length=100) | None = None
    warning: list[WarningItem] | None = None
    parent_name: dict[str, Any] | None = None
    read_only: bool | None = None
    original_id: uuid.UUID | None = None
    alert: list[uuid.UUID] | None = None
    contract: list[uuid.UUID] | None = None


class UserDailyLimit(BaseModel):
    class Config:
        extra = Extra.forbid

    point: conint(ge=0) | None = None
    document: conint(ge=0) | None = None
    log_row: conint(ge=0) | None = None
    traffic: conint(ge=0) | None = None
    iot_traffic: conint(ge=0) | None = None
    stream_traffic: conint(ge=0) | None = None
    file: conint(ge=0) | None = None
    request: conint(ge=0) | None = None
    request_time: conint(ge=0) | None = None


class NetworkDailyLimit(BaseModel):
    class Config:
        extra = Extra.forbid

    point: conint(ge=0) | None = None
    iot_traffic: conint(ge=0) | None = None
    request: conint(ge=0) | None = None
    request_time: conint(ge=0) | None = None


class DynamicLimitEnum(Enum):
    dynamic = 'dynamic'


class PointManagement(BaseModel):
    base_point:  conint(ge=0) | DynamicLimitEnum | None = None
    base_network:  conint(ge=0) | DynamicLimitEnum | None = None
    user_daily_limit: UserDailyLimit | None = None
    network_daily_limit: NetworkDailyLimit | None = None


class Subuser(BaseModel):
    class Config:
        extra = Extra.forbid

    first_name: str | None = None
    last_name: str | None = None
    email: constr(regex=r'^\S+@\S+\.\S+$') | None = None
    phone: str | None = None
    name: str | None = None
    role: str | None = None
    nickname: str | None = None
    language: str | None = None
    login_username: constr(regex=r'^\S+@\S+\.\S+$')
    session_token: uuid.UUID
    last_online: datetime.datetime | None = None
    meta: Meta = Field(..., title='meta-2.1')
    point_management: PointManagement | None = None



# generated by datamodel-codegen:
#   filename:  <stdin>
#   timestamp: 2023-02-17T08:37:52+00:00

from __future__ import annotations

import uuid
import datetime
from enum import Enum
from typing import Any, Literal

from pydantic import BaseModel, Extra, Field, conint, constr


class Engine(Enum):
    node = 'node'
    python = 'python'


class Executable(BaseModel):
    class Config:
        extra = Extra.forbid

    engine: Engine | None = None
    version: str | None = None


class SupportedFeature(Enum):
    foreground = 'foreground'
    background = 'background'
    widget = 'widget'


class Description(BaseModel):
    class Config:
        extra = Extra.forbid

    general: str | None = None
    foreground: str | None = None
    background: str | None = None
    widget: str | None = None
    version: str | None = None


class Status(Enum):
    idle = 'idle'
    commit = 'commit'
    revision = 'revision'
    publish = 'publish'
    uncommit = 'uncommit'
    unpublish = 'unpublish'
    republish = 'republish'
    disable = 'disable'
    reenable = 'reenable'
    duplicate = 'duplicate'
    pending = 'pending'
    uncommitted = 'uncommitted'
    published = 'published'
    not_updated = 'not updated'
    unpublished = 'unpublished'
    disabled = 'disabled'


class Installed(BaseModel):
    last_update: datetime.datetime | None = None
    quantity: int | None = None


class Assigned(Enum):
    unassigned = 'unassigned'
    admin = 'admin'




class WarningItem(BaseModel):
    message: str | None = None
    data: dict[str, Any] | None = None
    code: int | None = None


class Meta(BaseModel):
    class Config:
        extra = Extra.forbid

    id: uuid.UUID | None = None
    type: str | None = None
    version: constr(regex=r'(\d+)\.(\d+)') | None = None
    manufacturer: uuid.UUID | Assigned | None = None
    owner: uuid.UUID | Assigned | None = None
    created: datetime.datetime | Literal['not_available'] | None = None
    updated: datetime.datetime | Literal['not_available'] | None = None
    changed: datetime.datetime | Literal['not_available'] | None = None
    application: uuid.UUID | None = None
    revision: int | None = None
    trace: str | None = None
    oem: str | None = None
    deprecated: bool | None = None
    redirect: constr(regex=r'^[0-9a-zA-Z_-]+$', min_length=1, max_length=200) | None = None
    size: int | None = None
    path: str | None = None
    parent: uuid.UUID | None = None
    error: dict[str, Any] | None = None
    icon: str | None = None
    tag: list[constr(min_length=2, max_length=20)] | None = None
    tag_by_user: list[constr(min_length=2, max_length=20)] | None = None
    name_by_user: constr(max_length=100) | None = None
    warning: list[WarningItem] | None = None
    parent_name: dict[str, Any] | None = None
    read_only: bool | None = None
    original_id: uuid.UUID | None = None
    alert: list[uuid.UUID] | None = None
    contract: list[uuid.UUID] | None = None
    last_committed: uuid.UUID | None = None
    last_published: uuid.UUID | None = None


class Permission(BaseModel):
    create: list[str] | None = None
    permit_to_send_email: bool | None = None
    permit_to_send_sms: bool | None = None


class ConcernItem(BaseModel):
    class Config:
        extra = Extra.forbid

    concern: str
    line: conint(ge=0, multiple_of=1)


class ConcernFileItem(BaseModel):
    class Config:
        extra = Extra.forbid

    timestamp: datetime.datetime
    concern: list[ConcernItem] = Field(..., min_items=1)


class Meta1(BaseModel):
    class Config:
        extra = Extra.forbid

    id: uuid.UUID | None = None
    type: str | None = None
    version: constr(regex=r'(\d+)\.(\d+)') | None = None
    manufacturer: uuid.UUID | Assigned | None = None
    owner: uuid.UUID | Assigned | None = None
    created: datetime.datetime | Literal['not_available'] | None = None
    updated: datetime.datetime | Literal['not_available'] | None = None
    changed: datetime.datetime | Literal['not_available'] | None = None
    application: uuid.UUID | None = None
    revision: int | None = None
    trace: str | None = None
    oem: str | None = None
    deprecated: bool | None = None
    redirect: constr(regex=r'^[0-9a-zA-Z_-]+$', min_length=1, max_length=200) | None = None
    size: int | None = None
    path: str | None = None
    parent: uuid.UUID | None = None
    error: dict[str, Any] | None = None
    icon: str | None = None
    tag: list[constr(min_length=2, max_length=20)] | None = None
    tag_by_user: list[constr(min_length=2, max_length=20)] | None = None
    name_by_user: constr(max_length=100) | None = None
    warning: list[WarningItem] | None = None
    parent_name: dict[str, Any] | None = None
    read_only: bool | None = None
    original_id: uuid.UUID | None = None
    alert: list[uuid.UUID] | None = None
    contract: list[uuid.UUID] | None = None


class FileItem(BaseModel):
    class Config:
        extra = Extra.forbid

    name: str | None = None
    type: str | None = None
    size: conint(ge=0, multiple_of=1) | None = None
    published: bool | None = None
    concernFile: list[ConcernFileItem] | None = None
    meta: Meta1 = Field(..., title='meta-2.1')


class Version(BaseModel):
    class Config:
        extra = Extra.forbid

    name: str
    name_identifier: constr(regex=r'[0-9a-z_]') | None = None
    application: uuid.UUID | None = None
    author: str | None = None
    version_app: constr(regex=r'^(\d+\.\d+\.\d+)$') | None = None
    commit: str | None = None
    title: str | None = None
    session_user: bool | None = None
    concern: str | None = None
    native: bool | None = None
    name_folder: str | None = None
    icon: uuid.UUID | None = None
    priority: conint(ge=1, multiple_of=1) | None = None
    time_commit: datetime.datetime | None = None
    version_to_clone: uuid.UUID | None = None
    installation: list[uuid.UUID] | None = None
    uninstallable: bool | None = None
    published: bool | None = None
    executable: Executable | None = None
    max_number_installation: conint(ge=1, multiple_of=1) | None = None
    object_requested: dict[str, Any] | None = None
    supported_features: list[SupportedFeature] | None = None
    description: Description | None = None
    status: Status
    used_files: dict[str, Any]
    installed: Installed | None = None
    meta: Meta = Field(..., title='meta-2.1')
    permission: Permission | None = None
    file: list[FileItem | uuid.UUID] | None = None



from .create.acl import Acl
# from .create.add_not_owned import AddNotOwned
# from .create.admin_feature import AdminFeature
# from .create.alert import Alert
# from .create.analytics import Analytics
# from .create.analytics_model import AnalyticsModel
# from .create.application import Application
# from .create.application_product import ApplicationProduct
# from .create.azure import Azure
# from .create.bus_stream import BusStream
# from .create.cockroach import Cockroach
# from .create.console import Console
# from .create.couchbase import Couchbase
# from .create.creator import Creator
# from .create.customer import Customer
# from .create.customer_invoice import CustomerInvoice
# from .create.dashboard import Dashboard
# from .create.data import Data
# from .create.email_draft import EmailDraft
# from .create.email_send import EmailSend
# from .create.email_translation import EmailTranslation
# from .create.eventlog import Eventlog
# from .create.expected_state import ExpectedState
# from .create.extsync import Extsync
# from .create.extsync_response import ExtsyncResponse
# from .create.file import File
# from .create.filter import Filter
# from .create.firmware import Firmware
# from .create.fite import Fite
# from .create.homepage import Homepage
# from .create.installation import Installation
# from .create.mda import Mda
from .create.meta import Meta
# from .create.metrics import Metrics
# from .create.mqtt import Mqtt
# from .create.n1ql import N1ql
# from .create.notification import Notification
# from .create.oauth import Oauth
# from .create.oauth_client import OauthClient
# from .create.oauth_connect import OauthConnect
# from .create.oauth_external import OauthExternal
# from .create.ontology import Ontology
# from .create.parser import Parser
# from .create.permission import Permission
# from .create.permission_group import PermissionGroup
# from .create.permission_rule import PermissionRule
# from .create.point import Point
# from .create.price import Price
# from .create.product import Product
# from .create.promote import Promote
# from .create.prototype import Prototype
# from .create.purse import Purse
# from .create.recovery import Recovery
# from .create.recovery_cleanup import RecoveryCleanup
# from .create.recovery_repeat import RecoveryRepeat
# from .create.repeat import Repeat
# from .create.register import Register
# from .create.seller import Seller
from .create.session import Session
# from .create.sms_draft import SmsDraft
# from .create.sms_send import SmsSend
# from .create.status import Status
# from .create.stream import Stream
# from .create.subscription import Subscription
# from .create.subuser import Subuser
# from .create.tax import Tax

# from .create.transfer_billing import TransferBilling
# from .create.transfer_translation import TransferTranslation
# from .create.transfer_email import TransferEmail
# from .create.transfer_user import TransferUser

from .create.user import User
# from .create.version import Version
# from .create.virtualuser import Virtualuser
# from .create.websocket import Websocket
# from .create.widget import Widget

from .create.network import Network
from .create.network import Device
from .create.network import Value
from .create.network import State

__all__ = [
    'Network',
    'Device',
    'Value',
    'State',

    'Acl',
    # 'AddNotOwned',
    # 'AdminFeature',
    # 'Alert',
    # 'Analytics',
    # 'AnalyticsModel',
    # 'Application',
    # 'ApplicationProduct',
    # 'Azure',
    # 'BusStream',
    # 'Cockroach',
    # 'Console',
    # 'Couchbase',
    # 'Creator',
    # 'Customer',
    # 'CustomerInvoice',
    # 'Dashboard',
    # 'Data',
    # 'EmailDraft',
    # 'EmailSend',
    # 'EmailTranslation',
    # 'Eventlog',
    # 'ExpectedState',
    # 'Extsync',
    # 'ExtsyncResponse',
    # 'File',
    # 'Filter',
    # 'Firmware',
    # 'Fite',
    # 'Homepage',
    # 'Installation',
    # 'Mda',
    'Meta',
    # 'Metrics',
    # 'Mqtt',
    # 'N1ql',
    # 'Notification',
    # 'Oauth',
    # 'OauthClient',
    # 'OauthConnect',
    # 'OauthExternal',
    # 'Ontology',
    # 'Parser',
    # 'Permission',
    # 'PermissionGroup',
    # 'PermissionRule',
    # 'Point',
    # 'Price',
    # 'Product',
    # 'Promote',
    # 'Prototype',
    # 'Purse',
    # 'Recovery',
    # 'RecoveryCleanup',
    # 'Repeat',
    # 'Register',
    # 'RecoveryRepeat',
    # 'Seller',
    'Session',
    # 'SmsDraft',
    # 'SmsSend',
    # 'Status',
    # 'Stream',
    # 'Subscription',
    # 'Subuser',
    # 'Tax',
    # 'TransferBilling',
    # 'TransferTranslation',
    # 'TransferEmail',
    # 'TransferUser',
    'User',
    # 'Version',
    # 'Virtualuser',
    # 'Websocket',
    # 'Widget',
]



#!/usr/bin/env python

###############################################################################
#
# Wappsto Websocket
#
# Wappsto Websocket is used to get live change update.
#
#
#
# Link:
# https://documentation.dev.wappsto.com/#/docs/streaming/streaming
#
###############################################################################

import json
import logging
import httpx
import threading
import urllib
import uuid

from pydantic import parse_obj_as
from pydantic import ValidationError

from typing import Any
from typing import Callable
from typing import Dict
from typing import Union

# TODO: Use SlxJsonRpc for the Configuration messages. Instead of:

import websocket

from .schemas.base import WappstoService
from .schemas.websocket_schema import EventStreamSchema
from .schemas.websocket_schema import RPCRequest
from .schemas.websocket_schema import RPCSuccess
from .schemas.websocket_schema import HttpMethods 


name = "WappstoWS"
__version__ = "v0.5.0"


class WappstoWebSocket:
    def __init__(
        self,
        session=None,
        username=None,
        password=None,
        service="wappsto.com",
        parameters=None,
        full=True,
        version="2.1",
        timeout=None,
    ):
        self.log = logging.getLogger(__name__)
        self.log.addHandler(logging.NullHandler())
        # websocket.enableTrace(True)

        self.version = version
        self.url = f"{service}/services/{self.version}"
        self.parameters: dict

        self.open = False
        self.timeout = timeout

        if not session:
            if not username or not password:
                msg = "missing required positional argument(s): "
                msg += "'username' " if not username else ""
                msg += "'password' " if not password else ""
                raise TypeError(msg)
            session = self._create_session(username=username, password=password)
            if not session:
                msg = "An Error happened doing session creation."
                self.log.error(msg)
                raise ConnectionError(msg)

        if not isinstance(parameters, dict):
            self.parameters: dict = {}
        else:
            self.parameters = parameters

        self.parameters["X-session"] = session
        self.parameters["full"] = full

        # self.parameters['subscription'] = "[" + ",".join(sub_type) + "]"
        # websocket.enableTrace(True)

        # To keep track of the config msg.
        self._config_reply_wait: Dict[str, Any] = {}

        self.callback_list: Dict[
            uuid.UUID, Callable[[uuid.UUID, EventStreamSchema], None]
        ] = {}

        # self._unused_data = []

        self.ws = websocket.WebSocketApp(
            url=self._get_ws_link(),
            on_message=self.__on_message,
            on_error=self.__on_error,
            on_open=self.__on_open,
            # on_ping= self.__on_ping,
            # on_pong=self.__on_pong,
            # on_count=self.__on_count,
            on_close=self.__on_close,
        )
        self.wst = threading.Thread(target=self.ws.run_forever)
        # self.wst.daemon = True
        self.wst.start()

    def is_open(self):
        return self.open

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _config_wait(self, _id: str, data):
        """Send the config data, & wait for reply."""
        _temp_event = threading.Event()
        self._config_reply_wait[_id] = _temp_event
        self.ws.send(data=data.json())
        _temp_event.wait(timeout=self.timeout)
        return self._config_reply_wait.pop(_id)

    def _config_ready(self, _id: str, data):
        """Check if someone if waiting for config data & send it to them."""
        if _id in self._config_reply_wait:
            _temp_event = self._config_reply_wait.pop(_id)
            self._config_reply_wait[_id] = data
            _temp_event.set()
        else:
            print(f"Config reply: {data.id}")

    def _create_session(self, username, password) -> Union[Dict[str, uuid.UUID], None]:
        # Reuse the the one from Wappsto Rest.
        url = f"https://{self.url}/session"

        rdata = httpx.post(
            url=url,
            headers={"Content-type": "application/json"},
            json={"username": username, "password": password},
        )

        if rdata.status_code >= 300:  # TODO: Simpliefy me!
            try:
                error = json.loads(rdata.text)
            except TypeError:
                error = rdata.text
            self.log.error("Error on creating session:")
            self.log.error(error)
            return None
        rjson = json.loads(rdata.text)
        return rjson["meta"]["id"]

    def _get_ws_link(self) -> str:
        paras = "?" + urllib.parse.urlencode(self.parameters).lower()  # type: ignore
        link = f"wss://{self.url}/websocket/open{paras}"
        self.log.debug(f"Link: {link}")
        return link

    def subscribe(
        self,
        wappsto_type: WappstoService,
        unit_uuid: uuid.UUID,
        callback: Callable[[uuid.UUID, EventStreamSchema], None],
    ):
        """Subscribe a callback to a given UUID change."""
        if not isinstance(unit_uuid, uuid.UUID):
            unit_uuid = uuid.UUID(unit_uuid)
        # UNSURE: add lock?
        self.callback_list[unit_uuid] = callback

        data = RPCRequest(
            method=HttpMethods.PATCH,
            params={
                "url": f"/services/{self.version}/websocket/open/subscription",
                "data": f"/{wappsto_type.value}/{unit_uuid}",
            },
        )

        return self._config_wait(data.id, data)

    def unsubscribe(self, wappsto_type: WappstoService, unit_uuid: uuid.UUID):
        # # remove Subscription & callback
        if not isinstance(unit_uuid, uuid.UUID):
            unit_uuid = uuid.UUID(unit_uuid)

        data = RPCRequest(
            method=HttpMethods.DELETE,
            params={
                "url": f"/services/{self.version}/websocket/open/subscription",
                "data": f"/{wappsto_type.value}/{unit_uuid}",
            },
        )

        self._config_wait(data.id, data)

        return self.callback_list.pop(unit_uuid, None)

    def socket_info(self):
        # # Get Socket Info.
        data = RPCRequest(
            method=HttpMethods.GET,
            params={
                "url": f"/services/{self.version}/websocket/open",
            },
        )

        return self._config_wait(data.id, data)

    def __on_open(self, *args) -> None:
        self.log.debug(f"{self.url} open.")
        self.open = True

    def _default_cb(self, uuid: uuid.UUID, data):
        """Default Callback for the msg. Should not be called."""
        # self._unused_data.append(data)
        self.log.info(f"MSG: {data}")

    def __on_message(self, obj, message, *args) -> None:
        """Store message."""
        try:
            data = json.loads(message)
        except json.JSONDecodeError:
            print("Error Parsing data.")
            raise

        try:
            data = parse_obj_as(EventStreamSchema, data)
            cb_uuid = data.meta_object.id
        except ValidationError:
            try:
                data = parse_obj_as(RPCSuccess, data)
            except ValidationError:
                print(f"Error Parsing: {data}")
                raise
            else:
                self._config_ready(data.id, data)
                return

        self.log.debug(f"MSG received for: {cb_uuid}")

        # UNSURE: add lock?
        self.callback_list.get(cb_uuid, self._default_cb)(uuid.UUID(cb_uuid), data)

    def __on_error(self, exception, *args, **kwargs) -> None:
        self.log.error(f"exception: {exception}")

    def __on_ping(self, ws, data) -> None:
        self.log.debug("Ping")

    def __on_pong(self, ws, data) -> None:
        self.log.debug("Pong")

    def __on_data(self, ws, data, opcode, finish: bool) -> None:
        self.log.error(f"ws: {ws}")
        self.log.info(f"MSG: {data}")
        self.log.info(f"Code: {opcode}")

    def __on_count(self, ws, data, finish: bool) -> None:
        pass

    def __on_close(self, *args, **kwargs) -> None:
        self.log.info(f"Websocket: {self.url} was closed")

    def close(self):
        # NOTE: There is a 10sec timeout on close, because of the 10sec read timeout.
        self.ws.close()



# generated by datamodel-codegen:
#   filename:  <stdin>
#   timestamp: 2023-02-17T08:26:53+00:00

from __future__ import annotations

import uuid
import datetime
from enum import Enum
from typing import Any, Literal

from pydantic import BaseModel, Extra, Field, constr


class Level(Enum):
    founder = 'founder'
    admin = 'admin'
    sales = 'sales'
    empty = 'empty'
    custom = 'custom'


class Access(BaseModel):
    pages: list[str] | None = None
    tabs: list[str] | None = None


class Acl(BaseModel):
    change_manufacturer: bool | None = None
    change_owner: bool | None = None
    share_object: bool | None = None


class Admin(BaseModel):
    handle_admin: bool | None = None


class Application(BaseModel):
    create_admin_application: bool | None = None
    delete_published: bool | None = None
    disable_and_reenable: bool | None = None
    handle_application_product: bool | None = None
    handle_publishing: bool | None = None
    retrieve_oauth_secrets: bool | None = None


class Billing(BaseModel):
    change_user_subscription: bool | None = None
    handle_subscription: bool | None = None
    handle_tax: bool | None = None
    show_stripe: bool | None = None


class Communication(BaseModel):
    create_translation: bool | None = None
    delete_translation: bool | None = None
    publish_translation: bool | None = None
    send_email: bool | None = None
    send_notification: bool | None = None
    send_sms: bool | None = None
    update_email: bool | None = None
    update_sms: bool | None = None
    update_translation: bool | None = None


class Customer(BaseModel):
    handle_customer: bool | None = None
    handle_seller: bool | None = None
    list_customer: bool | None = None
    list_seller: bool | None = None


class Couchbase(BaseModel):
    access_user_bucket: bool | None = None
    delete: bool | None = None
    retrieve: bool | None = None
    save: bool | None = None
    retrieve_meta: bool | None = None


class Documentation(BaseModel):
    access_admin_documentation: bool | None = None


class Installation(BaseModel):
    restart_with_new_process: bool | None = None
    restart_with_new_user: bool | None = None
    retrieve_oauth_secret_token: bool | None = None
    stop_background: bool | None = None
    log_background: bool | None = None


class Assigned(Enum):
    unassigned = 'unassigned'
    admin = 'admin'


class WarningItem(BaseModel):
    message: str | None = None
    data: dict[str, Any] | None = None
    code: int | None = None


class Meta(BaseModel):
    class Config:
        extra = Extra.forbid

    id: uuid.UUID | None = None
    type: str | None = None
    version: constr(regex=r'(\d+)\.(\d+)') | None = None
    manufacturer: uuid.UUID | Assigned | None = None
    owner: uuid.UUID | Assigned | None = None
    created: datetime.datetime | Literal['not_available'] | None = None
    updated: datetime.datetime | Literal['not_available'] | None = None
    changed: datetime.datetime | Literal['not_available'] | None = None
    application: uuid.UUID | None = None
    revision: int | None = None
    trace: str | None = None
    oem: str | None = None
    deprecated: bool | None = None
    redirect: constr(regex=r'^[0-9a-zA-Z_-]+$', min_length=1, max_length=200) | None = None
    size: int | None = None
    path: str | None = None
    parent: uuid.UUID | None = None
    error: dict[str, Any] | None = None
    icon: str | None = None
    tag: list[constr(min_length=2, max_length=20)] | None = None
    tag_by_user: list[constr(min_length=2, max_length=20)] | None = None
    name_by_user: constr(max_length=100) | None = None
    warning: list[WarningItem] | None = None
    parent_name: dict[str, Any] | None = None
    read_only: bool | None = None
    original_id: uuid.UUID | None = None
    alert: list[uuid.UUID] | None = None
    contract: list[uuid.UUID] | None = None


class Micro(BaseModel):
    read_environment: bool | None = None
    retrieve_micro: bool | None = None
    show_environment: bool | None = None
    show_grafana: bool | None = None
    update_environment: bool | None = None


class Network(BaseModel):
    contact_directly: bool | None = None
    update_creator: bool | None = None
    handle_firmware: bool | None = None


class Recovery(BaseModel):
    recover_background: bool | None = None
    recover_cleanup: bool | None = None
    recover_repeat: bool | None = None


class Services(BaseModel):
    cockroach: bool | None = None
    full_search: bool | None = None
    n1ql: bool | None = None
    nats: bool | None = None


class Session(BaseModel):
    write_permission: bool | None = None


class Support(BaseModel):
    delete_historical: bool | None = None
    handle_cleanup_historical: bool | None = None
    handle_delayed_state: bool | None = None
    handle_pending: bool | None = None


class Transfer(BaseModel):
    promote_application: bool | None = None
    promote_email_and_sms: bool | None = None
    promote_subscription: bool | None = None
    promote_translation: bool | None = None
    transfer_user: bool | None = None


class Uiadmin(BaseModel):
    show_create_object_button: bool | None = None
    show_delete_object_button: bool | None = None
    show_update_object_button: bool | None = None


class User(BaseModel):
    ban_user: bool | None = None
    change_password: bool | None = None
    change_username: bool | None = None
    create_new_user: bool | None = None
    delete_user: bool | None = None
    reset_captcha: bool | None = None
    verify_communication: bool | None = None


class Wallet(BaseModel):
    change_points: bool | None = None
    correct_points: bool | None = None
    reset_points: bool | None = None


class AdminFeature(BaseModel):
    class Config:
        extra = Extra.allow

    level:  Level | None = None
    access:  Access | None = None
    acl:  Acl | None = None
    admin:  Admin | None = None
    application:  Application | None = None
    billing:  Billing | None = None
    communication:  Communication | None = None
    customer:  Customer | None = None
    couchbase:  Couchbase | None = None
    documentation:  Documentation | None = None
    installation:  Installation | None = None
    meta: Meta | None = Field(None, title='meta-2.1')
    micro:  Micro | None = None
    network: Network | None = None
    recovery:  Recovery | None = None
    services:  Services | None = None
    session:  Session | None = None
    support:  Support | None = None
    transfer:  Transfer | None = None
    uiadmin:  Uiadmin | None = None
    user:  User | None = None
    wallet:  Wallet | None = None



from ..path_helpers import Wrapper


class StateWarningItem:
    __path__ = 'device.value.state.meta.warning'
    overrite_list = [
        'message',
        'data',
        'code',
    ]
    for tmpop8yroywhfgkljsdyekj_name in overrite_list:
        locals()[tmpop8yroywhfgkljsdyekj_name] = Wrapper(
            tmpop8yroywhfgkljsdyekj_name, __path__
        )


class StateMeta:
    __path__ = 'device.value.state.meta'
    overrite_list = [
        'type',
        'id',
        'version',
        'manufacturer',
        'owner',
        'parent',
        'created',
        'updated',
        'changed',
        'application',
        'deprecated',
        'iot',
        'revision',
        'size',
        'path',
        'oem',
        'accept_manufacturer_as_owner',
        'redirect',
        'error',
        'trace',
        'set',
        'contract',
        'historical',
        'deletion',
    ]
    for tmpop8yroywhfgkljsdyekj_name in overrite_list:
        locals()[tmpop8yroywhfgkljsdyekj_name] = Wrapper(
            tmpop8yroywhfgkljsdyekj_name, __path__
        )

    warning = StateWarningItem


class State:
    __path__ = 'device.value.state'
    overrite_list = [
        'timestamp',
        'data',
        'type',
        'status',
    ]
    for tmpop8yroywhfgkljsdyekj_name in overrite_list:
        locals()[tmpop8yroywhfgkljsdyekj_name] = Wrapper(
            tmpop8yroywhfgkljsdyekj_name, __path__
        )

    meta = StateMeta


class ValueWarningItem:
    __path__ = 'device.value.meta.warning'
    overrite_list = [
        'message',
        'data',
        'code',
    ]
    for tmpop8yroywhfgkljsdyekj_name in overrite_list:
        locals()[tmpop8yroywhfgkljsdyekj_name] = Wrapper(
            tmpop8yroywhfgkljsdyekj_name, __path__
        )


class ValueMeta:
    __path__ = 'device.value.meta'
    overrite_list = [
        'type',
        'id',
        'version',
        'manufacturer',
        'owner',
        'parent',
        'created',
        'updated',
        'changed',
        'application',
        'deprecated',
        'iot',
        'revision',
        'size',
        'path',
        'oem',
        'accept_manufacturer_as_owner',
        'redirect',
        'error',
        'trace',
        'set',
        'contract',
        'historical',
        'deletion',
    ]
    for tmpop8yroywhfgkljsdyekj_name in overrite_list:
        locals()[tmpop8yroywhfgkljsdyekj_name] = Wrapper(
            tmpop8yroywhfgkljsdyekj_name, __path__
        )

    warning = ValueWarningItem


class ValueInfo:
    __path__ = 'device.value.info'
    overrite_list = [
        'enabled',
    ]
    for tmpop8yroywhfgkljsdyekj_name in overrite_list:
        locals()[tmpop8yroywhfgkljsdyekj_name] = Wrapper(
            tmpop8yroywhfgkljsdyekj_name, __path__
        )


class Number:
    __path__ = 'device.value.number'
    overrite_list = [
        'min',
        'max',
        'step',
        'mapping',
        'meaningful_zero',
        'ordered_mapping',
        'si_conversion',
        'unit',
    ]
    for tmpop8yroywhfgkljsdyekj_name in overrite_list:
        locals()[tmpop8yroywhfgkljsdyekj_name] = Wrapper(
            tmpop8yroywhfgkljsdyekj_name, __path__
        )


class String:
    __path__ = 'device.value.string'
    overrite_list = [
        'max',
        'encoding',
    ]
    for tmpop8yroywhfgkljsdyekj_name in overrite_list:
        locals()[tmpop8yroywhfgkljsdyekj_name] = Wrapper(
            tmpop8yroywhfgkljsdyekj_name, __path__
        )


class Blob:
    __path__ = 'device.value.blob'
    overrite_list = [
        'max',
        'encoding',
    ]
    for tmpop8yroywhfgkljsdyekj_name in overrite_list:
        locals()[tmpop8yroywhfgkljsdyekj_name] = Wrapper(
            tmpop8yroywhfgkljsdyekj_name, __path__
        )


class Xml:
    __path__ = 'device.value.xml'
    overrite_list = [
        'xsd',
        'namespace',
    ]
    for tmpop8yroywhfgkljsdyekj_name in overrite_list:
        locals()[tmpop8yroywhfgkljsdyekj_name] = Wrapper(
            tmpop8yroywhfgkljsdyekj_name, __path__
        )


class ValueEventlogWarningItem:
    __path__ = 'device.value.eventlog.meta.warning'
    overrite_list = [
        'message',
        'data',
        'code',
    ]
    for tmpop8yroywhfgkljsdyekj_name in overrite_list:
        locals()[tmpop8yroywhfgkljsdyekj_name] = Wrapper(
            tmpop8yroywhfgkljsdyekj_name, __path__
        )


class ValueEventlogMeta:
    __path__ = 'device.value.eventlog.meta'
    overrite_list = [
        'type',
        'id',
        'version',
        'manufacturer',
        'owner',
        'parent',
        'created',
        'updated',
        'changed',
        'application',
        'deprecated',
        'iot',
        'revision',
        'size',
        'path',
        'oem',
        'accept_manufacturer_as_owner',
        'redirect',
        'error',
        'trace',
        'set',
        'contract',
        'historical',
        'deletion',
        'icon',
        'alert',
    ]
    for tmpop8yroywhfgkljsdyekj_name in overrite_list:
        locals()[tmpop8yroywhfgkljsdyekj_name] = Wrapper(
            tmpop8yroywhfgkljsdyekj_name, __path__
        )

    warning = ValueEventlogWarningItem


class ValueEventlogItem:
    __path__ = 'device.value.eventlog'
    overrite_list = [
        'message',
        'timestamp',
        'info',
        'level',
        'type',
    ]

    for tmpop8yroywhfgkljsdyekj_name in overrite_list:
        locals()[tmpop8yroywhfgkljsdyekj_name] = Wrapper(
            tmpop8yroywhfgkljsdyekj_name, __path__
        )

    meta = ValueEventlogMeta


class Value:
    __path__ = 'device.value'
    overrite_list = [
        'name',
        'type',
        'description',
        'period',
        'delta',
        'permission',
        'status',
    ]
    for tmpop8yroywhfgkljsdyekj_name in overrite_list:
        locals()[tmpop8yroywhfgkljsdyekj_name] = Wrapper(
            tmpop8yroywhfgkljsdyekj_name, __path__
        )

    meta = ValueMeta
    state = State
    eventlog = ValueEventlogItem
    info = ValueInfo
    string = String
    number = Number
    blob = Blob
    xml = Xml


class DeviceGeo:
    __path__ = 'device.meta.geo'
    overrite_list = [
        'latitude',
        'longitude',
        'display_name',
        'address',
    ]
    for tmpop8yroywhfgkljsdyekj_name in overrite_list:
        locals()[tmpop8yroywhfgkljsdyekj_name] = Wrapper(
            tmpop8yroywhfgkljsdyekj_name, __path__
        )


class DeviceWarningItem:
    __path__ = 'device.meta.warning'
    overrite_list = [
        'message',
        'data',
        'code',
    ]
    for tmpop8yroywhfgkljsdyekj_name in overrite_list:
        locals()[tmpop8yroywhfgkljsdyekj_name] = Wrapper(
            tmpop8yroywhfgkljsdyekj_name, __path__
        )


class DeviceMeta:
    __path__ = 'device.meta'
    overrite_list = [
        'type',
        'id',
        'version',
        'manufacturer',
        'owner',
        'parent',
        'created',
        'updated',
        'changed',
        'application',
        'deprecated',
        'iot',
        'revision',
        'size',
        'path',
        'oem',
        'accept_manufacturer_as_owner',
        'redirect',
        'error',
        'trace',
        'set',
        'contract',
        'historical',
        'deletion',
    ]
    for tmpop8yroywhfgkljsdyekj_name in overrite_list:
        locals()[tmpop8yroywhfgkljsdyekj_name] = Wrapper(
            tmpop8yroywhfgkljsdyekj_name, __path__
        )

    warning = DeviceWarningItem

    geo = DeviceGeo


class StatusMeta:
    __path__ = 'device.status.meta'
    overrite_list = [
        'type',
        'id',
        'version',
        'manufacturer',
        'owner',
        'parent',
        'created',
        'updated',
        'changed',
        'application',
        'deprecated',
        'iot',
        'revision',
        'size',
        'path',
        'oem',
        'accept_manufacturer_as_owner',
        'redirect',
        'error',
        'trace',
        'set',
        'contract',
        'historical',
        'deletion',
        'icon',
        'alert',
    ]
    for tmpop8yroywhfgkljsdyekj_name in overrite_list:
        locals()[tmpop8yroywhfgkljsdyekj_name] = Wrapper(
            tmpop8yroywhfgkljsdyekj_name, __path__
        )

    warning = ValueEventlogWarningItem


class Status:
    __path__ = 'device.status'
    overrite_list = [
        'message',
        'timestamp',
        'data',
        'type',
        'level',
    ]
    for tmpop8yroywhfgkljsdyekj_name in overrite_list:
        locals()[tmpop8yroywhfgkljsdyekj_name] = Wrapper(
            tmpop8yroywhfgkljsdyekj_name, __path__
        )

    meta = StatusMeta


class DeviceInfo:
    __path__ = 'device.info'
    overrite_list = [
        'enabled',
    ]
    for tmpop8yroywhfgkljsdyekj_name in overrite_list:
        locals()[tmpop8yroywhfgkljsdyekj_name] = Wrapper(
            tmpop8yroywhfgkljsdyekj_name, __path__
        )


class Device:
    __path__ = 'device'
    overrite_list = [
        'name',
        'control_timeout',
        'control_when_offline',
        'manufacturer',
        'product',
        'version',
        'serial',
        'description',
        'protocol',
        'communication',
        'included',
        'inclusion_status',
        'firmware_status',
        'firmware_upload_progress',
        'firmware_available_version',
        'command',
    ]
    for tmpop8yroywhfgkljsdyekj_name in overrite_list:
        locals()[tmpop8yroywhfgkljsdyekj_name] = Wrapper(
            tmpop8yroywhfgkljsdyekj_name, __path__
        )

    meta = DeviceMeta
    status = Status
    value = Value
    info = DeviceInfo



# generated by datamodel-codegen:
#   filename:  <stdin>
#   timestamp: 2023-02-17T08:29:30+00:00

from __future__ import annotations

from enum import Enum
from typing import Any
from typing import Literal

from pydantic import BaseModel, Extra, Field


class WappstoVersion(Enum):
    V2_0 = "2.0"
    V2_1 = "2.1"


class WarningItem(BaseModel):
    message: str | None = None
    data: dict[str, Any] | None = None
    code: int | None = None


class Meta(BaseModel):
    class Config:
        extra = Extra.forbid

    type: Literal['deletelist']
    version: WappstoVersion | None = None
    trace: str | None = None
    oem: str | None = None
    deprecated: bool | None = None
    warning: list[WarningItem] | None = None


class UnsuccessfulItem(BaseModel):
    id: str | None = None
    message: str | None = None
    data: dict[str, Any] | None = None
    code: int | None = None


class Deletelist(BaseModel):
    class Config:
        extra = Extra.forbid

    deleted: list[str] | None = None
    child_deleted: list[str] | None = None
    shared_deleted: list[str] | None = None
    shared_child_deleted: list[str] | None = None
    not_deleted: list[str] | None = None
    child_not_deleted: list[str] | None = None
    shared_not_deleted: list[str] | None = None
    shared_child_not_deleted: list[str] | None = None
    pending: list[str] | None = None
    child_pending: list[str] | None = None
    owned_deleted: list[str] | None = None
    owned_not_deleted_because_it_has_parent: list[str] | None = None
    shared_not_deleted_because_it_has_parent: list[str] | None = None
    code: int
    message: str
    error: str | None = None
    internal_code: str | None = None
    meta: Meta = Field(..., title='meta_list-2.1')
    unsuccessful: list[UnsuccessfulItem] | None = None



# generated by datamodel-codegen:
#   filename:  <stdin>
#   timestamp: 2023-02-17T08:34:06+00:00

from __future__ import annotations

import uuid
from enum import Enum

from pydantic import BaseModel, Extra, Field, conint, constr


class UserEnum(Enum):
    global_ = 'global'


class IdEnum(Enum):
    global_ = 'global'


class PermissionRule(BaseModel):
    class Config:
        extra = Extra.forbid

    user: list[uuid.UUID | UserEnum] | None = None
    deeper: bool | None = None
    id: list[uuid.UUID | IdEnum] | None = None
    global_: bool | None = Field(None, alias='global')
    from_: bool | None = Field(None, alias='from')
    to: bool | None = None
    type: str | None = None
    expand: conint(ge=0, multiple_of=1) | None = None
    offset: conint(ge=0, multiple_of=1) | None = None
    limit: conint(ge=0, multiple_of=1) | None = None
    version: list[constr(regex=r'\d+(\.\d+)*')] | None = None
    verbose: bool | None = None



import datetime
import random
import string
import uuid

from enum import Enum

from typing import Any

from pydantic import BaseModel
from pydantic import Extra
from pydantic import validator

from .output.network import Network
from .output.network import Device
from .output.network import Value
from .output.network import State

from .base import BaseMeta
from .base import WappstoVersion


class EventStreamType(str, Enum):
    eventstream = "eventstream"


class MetaEventSchema(BaseModel):
    id: uuid.UUID
    type: EventStreamType
    version: WappstoVersion


class StreamEvents(str, Enum):
    CREATE = "create"
    UPDATE = "update"
    DELETE = "delete"
    DIRECT = "direct"


class HttpMethods(str, Enum):
    PATCH = "PATCH"
    GET = "GET"
    PUT = "PUT"
    POST = "POST"
    DELETE = "DELETE"


class RpcVersion(str, Enum):
    v2_0 = "2.0"


class EventStreamSchema(BaseModel):
    data: Network | Device | Value | State
    event: StreamEvents
    meta: MetaEventSchema
    meta_object: BaseMeta
    path: str
    timestamp: datetime.datetime

    # @validator('path')
    # def path_check(cls, v, values, **kwargs):
    #     for selftype, selfid in parwise(v.split("/")[1:]):
    #         WappstoMetaType(selftype)
    #         lasttype = selftype
    #         if selfid:
    #             uuid.UUID(selfid)
    #     if "meta_object" in values and "type" in values["meta_object"]:
    #         if values["meta_object"].type != lasttype:
    #             raise ValueError('Path do not match Type')
    #     return v


__session_count: int = 0
__session_id: str = "".join(random.choices(string.ascii_letters + string.digits, k=10))


def _id_gen():
    """Create an unique Rpc-id."""
    global __session_count
    global __session_id
    __session_count += 1
    return f"{__session_id}_WSS_CONFIGS_{__session_count}"


class RPCRequest(BaseModel):
    params: Any | None
    method: HttpMethods
    jsonrpc: RpcVersion | None = RpcVersion.v2_0
    id: str | int | None = None

    class Config:
        extra = Extra.forbid

    @validator("id", pre=True, always=True)
    def id_autofill(cls, v):
        return v or _id_gen()


class RPCSuccess(BaseModel):
    result: Any
    jsonrpc: RpcVersion | None = RpcVersion.v2_0
    id: str | int | None = None

    class Config:
        extra = Extra.forbid

    @validator("id", pre=True, always=True)
    def id_autofill(cls, v):
        return v or _id_gen()



# generated by datamodel-codegen:
#   filename:  <stdin>
#   timestamp: 2023-02-17T08:29:52+00:00

from __future__ import annotations

import uuid
import datetime
from enum import Enum
from typing import Any, Literal

from pydantic import BaseModel, Extra, Field, conint, constr


class MethodEnum(Enum):
    retrieve = 'retrieve'
    update = 'update'
    add = 'add'
    delete = 'delete'


class Conjunction(Enum):
    and_ = 'and'
    or_ = 'or'


class WrongObject(Enum):
    expanded = 'expanded'
    compact = 'compact'
    hidden = 'hidden'


class Schema(Enum):
    create = 'create'
    update = 'update'
    output = 'output'
    query = 'query'


class AclAttribute(Enum):
    parent_name = 'parent_name'


class HistoricalStateDataOperationEnum(Enum):
    avg = 'avg'
    mean = 'mean'
    exists = 'exists'
    count = 'count'
    geometric_mean = 'geometric_mean'
    max = 'max'
    min = 'min'
    sqrdiff = 'sqrdiff'
    stddev = 'stddev'
    sum = 'sum'
    variance = 'variance'
    harmonic_mean = 'harmonic_mean'
    first = 'first'
    last = 'last'
    count_distinct = 'count_distinct'
    median = 'median'
    lower_quartile = 'lower_quartile'
    upper_quartile = 'upper_quartile'


class HistoricalStateDataCompareEnum(Enum):
    bigger_and_equal = 'bigger_and_equal'
    bigger = 'bigger'
    smaller_and_equal = 'smaller_and_equal'
    smaller = 'smaller'
    different = 'different'
    equal = 'equal'
    in_range = 'in_range'
    outside_range = 'outside_range'


class HistoricalStateDataGroupBy(Enum):
    year = 'year'
    quarter = 'quarter'
    month = 'month'
    week = 'week'
    day = 'day'
    hour = 'hour'
    minute = 'minute'
    second = 'second'
    millisecond = 'millisecond'
    microsecond = 'microsecond'


class Device(BaseModel):
    class Config:
        extra = Extra.forbid

    id: list[uuid.UUID] | None = None
    offset: conint(ge=0, multiple_of=1) | None = None
    limit: conint(ge=0, multiple_of=1) | None = None
    manufacturer: bool | None = None
    owned: bool | None = None
    accepted: bool | None = None
    shared_with: list[uuid.UUID] | None = None
    not_shared_with: list[uuid.UUID] | None = None
    has_parent: bool | None = None
    version: list[constr(regex=r'\d+(\.\d+)*')] | None = None
    method: list[MethodEnum] | None = None
    deep: conint(ge=1, multiple_of=1) | None = None
    global_: bool | None = Field(None, alias='global')
    full_search: bool | None = None
    from_last: bool | None = None
    order_by: str | None = None
    permission_group: list[Literal['all'] | uuid.UUID] | None = None
    clean: bool | None = None
    conjunction:  Conjunction | None = None
    count: bool | None = None
    asynchronous: bool | uuid.UUID | None = None
    propagate: bool | None = None
    acl_code: uuid.UUID | None = None
    expand: conint(ge=0, multiple_of=1) | None = None
    quantity: int | Literal['all'] | None = None
    message: str | None = None
    opt_attr: list[str] | None = None
    opt_value: list[str] | None = None
    verbose: bool | list[str] | None = None
    compact: list[str] | None = None
    fetch: bool | None = None
    go_internal: bool | None = None
    expand_from_bottom: bool | None = None
    wrong_object:  WrongObject | None = None
    cache: conint(ge=1, le=86400) | None = None
    cache_refresh: bool | None = None
    csv_name: str | None = None
    csv_separator: str | None = None
    prev_notification: uuid.UUID | None = None
    schema_: Schema | None = Field(None, alias='schema')
    test_mode: bool | None = None
    acl_attributes: list[AclAttribute] | None = None
    historical_device_geo_rectangular: list[list[Any]] | None = Field(None, max_items=2, min_items=2)
    historical_device_geo_start: datetime.datetime | None = None
    historical_device_geo_end: datetime.datetime | None = None
    historical_network_iot_was_online: bool | None = None
    historical_network_iot_was_offline: bool | None = None
    historical_network_iot_start: datetime.datetime | None = None
    historical_network_iot_end: datetime.datetime | None = None
    historical_state_data_operation: list[HistoricalStateDataOperationEnum] | None = None
    historical_state_data_compare: list[HistoricalStateDataCompareEnum] | None = None
    historical_state_data_value: list | None = None
    historical_state_data_group_by:  HistoricalStateDataGroupBy | None = None
    historical_state_data_start: datetime.datetime | None = None
    historical_state_data_end: datetime.datetime | None = None
    geo_polygon: list[list[list[Any]]] | None = None
    geo_rectangular: list[list[Any]] | None = Field(
        None, max_items=2, min_items=2
    )
    geo_country: list[str] | None = None
    geo_city: list[str] | None = None
    geo_country_code: list[str] | None = None
    table: bool | None = None
    refresh_geo: bool | None = None



# generated by datamodel-codegen:
#   filename:  <stdin>
#   timestamp: 2023-02-17T08:29:53+00:00

from __future__ import annotations

from pydantic import BaseModel, Extra


class Download(BaseModel):
    pass

    class Config:
        extra = Extra.forbid
